"""
Boticord API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the Boticord API.
:copyright: (c) 2022 Marakarka
:license: MIT, see LICENSE for more details.
"""

__title__ = 'boticordpy'
__author__ = 'Marakarka'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022 Marakarka'
__version__ = '2.0.2'

from .client import BoticordClient
from .webhook import Webhook

from .types import *



import asyncio
import typing

from .types import BumpResponse, CommentResponse

from aiohttp import web
import aiohttp


class Webhook:
    """Represents a client that can be used to work with Boticord Webhooks.
    IP of the server - your machine IP. (`0.0.0.0`)

    Args:
        x_hook_key (:obj:`str`)
            X-hook-key to check the auth of incoming request.
        endpoint_name (:obj:`str`)
            Name of endpoint (for example: `/bot`)

    Keyword Arguments:
        loop: `asyncio loop`
    """
    __slots__ = (
        "_webserver",
        "_listeners",
        "_is_running",
        "__app",
        "_endpoint_name",
        "_x_hook_key",
        "_loop"
    )

    __app: web.Application
    _webserver: web.TCPSite

    def __init__(self, x_hook_key: str, endpoint_name: str, **kwargs) -> None:
        self._x_hook_key = x_hook_key
        self._endpoint_name = endpoint_name
        self._listeners = {}
        self.__app = web.Application()
        self._is_running = False
        self._loop = kwargs.get('loop') or asyncio.get_event_loop()

    def listener(self, response_type: str):
        """Decorator to set the listener.
        Args:
            response_type (:obj:`str`)
                Type of response (Check reference page)
        """
        def inner(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
            self._listeners[response_type] = func
            return func

        return inner

    def register_listener(self, response_type: str, callback: typing.Any):
        """Method to set the listener.
        Args:
            response_type (:obj:`str`)
                Type of response (Check reference page)
            callback (:obj:`function`)
                Coroutine Callback Function
        """
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")

        self._listeners[response_type] = callback
        return self

    async def _interaction_handler(self, request: aiohttp.web.Request) -> web.Response:
        """Interaction handler"""
        auth = request.headers.get("X-Hook-Key")

        if auth == self._x_hook_key:
            data = await request.json()

            responder = self._listeners.get(data["type"])

            if responder is not None:
                await responder(
                    (BumpResponse if data["type"].endswith("_bump") else CommentResponse)(**data)
                )

            return web.Response(status=200)

        return web.Response(status=401)

    async def _run(self, port):
        self.__app.router.add_post("/" + self._endpoint_name, self._interaction_handler)

        runner = web.AppRunner(self.__app)
        await runner.setup()

        self._webserver = web.TCPSite(runner, "0.0.0.0", port)
        await self._webserver.start()

        self._is_running = True

    def start(self, port: int) -> None:
        """Method to start the webhook server

        Args:
            port (:obj:`int`)
                Port to start the webserver
        """
        self._loop.create_task(self._run(port))

    @property
    def is_running(self) -> bool:
        """If the server running?"""
        return self._is_running

    @property
    def listeners(self) -> dict:
        """Dictionary of listeners (`type`: `callback function`)"""
        return self._listeners

    @property
    def app(self) -> web.Application:
        """Web application that handles incoming requests"""
        return self.__app

    async def close(self) -> None:
        """Stop the webhooks server"""
        await self._webserver.stop()

        self._is_running = False



import asyncio

import aiohttp

from . import exceptions


class HttpClient:
    """
    Represents an HTTP client sending HTTP requests to the Top.gg API.

    Args:
        token (:obj:`str`)
            Your bot's Boticord API Token.

    Keyword Arguments:
        session: `aiohttp session`_
            The `aiohttp session`_ used for requests to the API.
        loop: `asyncio loop`
    """

    def __init__(self, auth_token, **kwargs):
        self.token = auth_token
        self.API_URL = "https://api.boticord.top/v1/"

        loop = kwargs.get('loop') or asyncio.get_event_loop()

        self.session = kwargs.get('session') or aiohttp.ClientSession(loop=loop)

    async def make_request(self,
                           method: str,
                           endpoint: str,
                           **kwargs):
        """Send requests to the API"""

        kwargs["headers"] = {
            "Content-Type": "application/json",
            "Authorization": self.token
        }

        url = f"{self.API_URL}{endpoint}"

        async with self.session.request(method,
                                        url,
                                        **kwargs) as response:
            data = await response.json()

            if response.status == 200:
                return data
            elif response.status == 401:
                raise exceptions.Unauthorized(response)
            elif response.status == 403:
                raise exceptions.Forbidden(response)
            elif response.status == 404:
                raise exceptions.NotFound(response)
            elif response.status == 429:
                raise exceptions.ToManyRequests(response)
            elif response.status == 500:
                raise exceptions.ServerError(response)
            elif response.status == 503:
                raise exceptions.ServerError(response)

        raise exceptions.HTTPException(response)

    def get_bot_info(self, bot_id: int):
        """Get information about the specified bot"""
        return self.make_request("GET", f"bot/{bot_id}")

    def get_bot_comments(self, bot_id: int):
        """Get list of specified bot comments"""
        return self.make_request("GET", f"bot/{bot_id}/comments")

    def post_bot_stats(self, stats: dict):
        """Post bot's stats"""
        return self.make_request("POST", "stats", json=stats)

    def get_server_info(self, server_id: int):
        """Get information about specified server"""
        return self.make_request("GET", f"server/{server_id}")

    def get_server_comments(self, server_id: int):
        """Get list of specified server comments"""
        return self.make_request("GET", f"server/{server_id}/comments")

    def post_server_stats(self, payload: dict):
        """Post server's stats"""
        return self.make_request("POST", "server", json=payload)

    def get_user_info(self, user_id: int):
        """Get information about the user"""
        return self.make_request("GET", f"profile/{user_id}")

    def get_user_comments(self, user_id: int):
        """Get specified user's comments"""
        return self.make_request("GET", f"user/{user_id}/comments")

    def get_user_bots(self, user_id: int):
        """Get bots of specified user"""
        return self.make_request("GET", f"bots/{user_id}")



import typing

KT = typing.TypeVar("KT")
VT = typing.TypeVar("VT")


def parse_response_dict(input_data: dict) -> dict:
    data = input_data.copy()

    for key, value in data.copy().items():
        converted_key = "".join(
            ["_" + x.lower() if x.isupper() else x for x in key]
        ).lstrip("_")

        if key != converted_key:
            del data[key]

        data[converted_key] = value

    return data


def parse_with_information_dict(bot_data: dict) -> dict:
    data = bot_data.copy()

    for key, value in data.copy().items():
        if key.lower() == "links":
            converted_key = "page_links"
        else:
            converted_key = "".join(
                ["_" + x.lower() if x.isupper() else x for x in key]
            ).lstrip("_")

        if key != converted_key:
            del data[key]

        if key.lower() == "information":
            for information_key, information_value in value.copy().items():
                converted_information_key = "".join(
                    ["_" + x.lower() if x.isupper() else x for x in information_key]
                ).lstrip("_")

                data[converted_information_key] = information_value

            del data["information"]
        else:
            data[converted_key] = value

    return data


def parse_user_comments_dict(response_data: dict) -> dict:
    data = response_data.copy()

    for key, value in data.copy().items():
        data[key] = [SingleComment(**comment) for comment in value]

    return data


class ApiData(dict, typing.MutableMapping[KT, VT]):
    """Base class used to represent received data from the API.
    """

    def __init__(self, **kwargs: VT) -> None:
        super().__init__(**parse_response_dict(kwargs))
        self.__dict__ = self


class SingleComment(ApiData):
    """This model represents single comment"""

    user_id: str
    """Comment's author Id (`str`)"""

    text: str
    """Comment content"""

    vote: int
    """Comment vote value (`-1,` `0`, `1`)"""

    is_updated: bool
    """Was comment updated?"""

    created_at: int
    """Comment Creation date timestamp"""

    updated_at: int
    """Last edit date timestamp"""

    def __init__(self, **kwargs):
        super().__init__(**parse_response_dict(kwargs))


class Bot(ApiData):
    """This model represents a bot, returned from the Boticord API"""
    id: str
    """Bot's Id"""

    short_code: typing.Optional[str]
    """Bot's page short code"""

    page_links: list
    """List of bot's page urls"""

    server: dict
    """Bot's support server"""

    bumps: int
    """Bumps count"""

    added: str
    """How many times users have added the bot?"""

    prefix: str
    """Bot's commands prefix"""

    permissions: int
    """Bot's permissions"""

    tags: list
    """Bot's search-tags"""

    developers: list
    """List of bot's developers Ids"""

    links: typing.Optional[dict]
    """Bot's social medias"""

    library: typing.Optional[str]
    """Bot's library"""

    short_description: typing.Optional[str]
    """Bot's short description"""

    long_description: typing.Optional[str]
    """Bot's long description"""

    badge: typing.Optional[int]
    """Bot's badge"""

    stats: dict
    """Bot's stats"""

    status: str
    """Bot's approval status"""

    def __init__(self, **kwargs):
        super().__init__(**parse_with_information_dict(kwargs))


class Server(ApiData):
    """This model represents a server, returned from the Boticord API"""

    id: str
    """Server's Id"""

    short_code: typing.Optional[str]
    """Server's page short code"""

    status: str
    """Server's approval status"""

    page_links: list
    """List of server's page urls"""

    bot: dict
    """Bot where this server is used for support users"""

    name: str
    """Name of the server"""

    avatar: str
    """Server's avatar"""

    members: list
    """Members counts - `[all, onlinw]`"""

    owner: typing.Optional[str]
    """Server's owner Id"""

    bumps: int
    """Bumps count"""

    tags: list
    """Server's search-tags"""

    links: dict
    """Server's social medias"""

    short_description: typing.Optional[str]
    """Server's short description"""

    long_description: typing.Optional[str]
    """Server's long description"""

    badge: typing.Optional[str]
    """Server's badge"""

    def __init__(self, **kwargs):
        super().__init__(**parse_with_information_dict(kwargs))


class UserProfile(ApiData):
    """This model represents profile of user, returned from the Boticord API"""

    id: str
    """Id of User"""

    status: str
    """Status of user"""

    badge: typing.Optional[str]
    """User's badge"""

    short_code: typing.Optional[str]
    """User's profile page short code"""

    site: typing.Optional[str]
    """User's website"""

    vk: typing.Optional[str]
    """User's VK Profile"""

    steam: typing.Optional[str]
    """User's steam account"""

    youtube: typing.Optional[str]
    """User's youtube channel"""

    twitch: typing.Optional[str]
    """User's twitch channel"""

    git: typing.Optional[str]
    """User's github/gitlab (or other git-service) profile"""

    def __init__(self, **kwargs):
        super().__init__(**parse_response_dict(kwargs))


class UserComments(ApiData):
    """This model represents all the user's comments on every page"""

    bots: list
    """Data from `get_bot_comments` method"""

    servers: list
    """Data from `get_server_comments` method"""

    def __init__(self, **kwargs):
        super().__init__(**parse_user_comments_dict(kwargs))


class SimpleBot(ApiData):
    """This model represents a short bot information (`id`, `short`).
    After that you can get more information about it using `get_bot_info` method."""

    id: str
    """Bot's Id"""

    short_code: typing.Optional[str]
    """Bot's page short code"""

    def __init__(self, **kwargs):
        super().__init__(**parse_response_dict(kwargs))


class CommentData(ApiData):
    """This model represents comment data (from webhook response)"""

    vote: dict
    """Comment vote data"""

    old: typing.Optional[str]
    """Old content of the comment"""

    new: typing.Optional[str]
    """New content of the comment"""

    def __init__(self, **kwargs):
        super().__init__(**parse_response_dict(kwargs))


def parse_webhook_response_dict(webhook_data: dict) -> dict:
    data = webhook_data.copy()

    for key, value in data.copy().items():
        if key.lower() == "data":
            for data_key, data_value in value.copy().items():
                if data_key == "comment":
                    data[data_key] = CommentData(**data_value)
                else:
                    converted_data_key = "".join(
                        ["_" + x.lower() if x.isupper() else x for x in data_key]
                    ).lstrip("_")

                    data[converted_data_key] = data_value

            del data["data"]
        else:
            data[key] = value

    return data


class BumpResponse(ApiData):
    """This model represents a webhook response (`bot bump`)."""

    type: str
    """Type of response (`bump`)"""

    user: str
    """Id of user who did the action"""

    at: int
    """Timestamp of the action"""

    def __init__(self, **kwargs):
        super().__init__(**parse_webhook_response_dict(kwargs))


class CommentResponse(ApiData):
    """This model represents a webhook response (`comment`)."""

    type: str
    """Type of response (`comment`)"""

    user: str
    """Id of user who did the action"""

    comment: CommentData
    """Information about the comment"""

    reason: typing.Optional[str]
    """Is comment deleted? so, why?"""

    at: int
    """Timestamp of the  action"""

    def __init__(self, **kwargs):
        super().__init__(**parse_webhook_response_dict(kwargs))

